#!/bin/bash

# Version 26/04/2012
# $Revision: 1.24 $
# By UDA'Software 2011

# Modulo per la gestione di una connessione ad un Server IRC.

#####################################################################
#____________________________________________________________________
XBSPSE_Crawler_RANDOM=${RANDOM}
CRAWLER_TIMEOUT_COUNTDOWN=0
LIST_COMMAND='!list'
MAX_LOAD_AVERAGE=1000

#####################################################################
#____________________________________________________________________
XBSPSE_Crawler_log_file=/tmp/XBSPSE_Crawler.log
XBSPSE_Crawler_dat_file=/tmp/XBSPSE_Crawler_${XBSPSE_Crawler_RANDOM}.dat
XBSPSE_Crawler_time_file=/tmp/XBSPSE_Crawler_${XBSPSE_Crawler_RANDOM}.time
XBSPSE_Crawler_lock_file=/tmp/XBSPSE_Crawler.lock

#####################################################################
#____________________________________________________________________
CODE_ERR_NO_HOST=2
CODE_ERR_NO_CHANNEL=3
CODE_ERR_CONNECTION=4
CODE_ERR_NO_DELETE_DAT_FILE=5
CODE_ERR_TIMEOUT=6
CODE_ERR_NO_QUEUE_LOCKED_FROM_THIS_PROCESS=7
CODE_ERR_LOST_LOCK_ON_QUEUE=8

#####################################################################
#____________________________________________________________________
function check_system_load_average() {
    current_load_average=$(uptime \
	| grep -o '[0-9]*\.[0-9]*' \
	| head -n1 \
	| sed 's@\.@@g' \
	| sed 's@^0*@@g')
    
    if [ ${current_load_average:-0} -gt ${MAX_LOAD_AVERAGE:-500} ]; then
	printf 'XBSPSE Crawler [%4d]: IL CARICO DEL SISTEMA (%d%%) SUPERA QUELLO CONSENTITO (%d%%), DEVO ABORTIRE.\n' ${CRAWLER_TIMEOUT_COUNTDOWN:-0} ${current_load_average:-0} ${MAX_LOAD_AVERAGE:-500} >&2
	
	CRAWLER_TIMEOUT_COUNTDOWN=0
    fi
}

function queue_connection_manager() {
    sleep_step_seconds=4
    
    case ${1:-GET} in
	"GET")
	    current_time=$(date +%s)
	    
	    if [ -f "${XBSPSE_Crawler_lock_file}" ]; then
		while [ ${CRAWLER_TIMEOUT_COUNTDOWN} -gt 0 ]; do
		    sleep ${sleep_step_seconds}
		    CRAWLER_TIMEOUT_COUNTDOWN=$((${CRAWLER_TIMEOUT_COUNTDOWN} - ${sleep_step_seconds}))
		    [ ${CRAWLER_TIMEOUT_COUNTDOWN} -lt 0 ] && CRAWLER_TIMEOUT_COUNTDOWN=0
		    
		    if [ ! -f "${XBSPSE_Crawler_lock_file}" ]; then
			queue_connection_manager 'GET'
			return $?
		    else
			lock_process_id=$(cat "${XBSPSE_Crawler_lock_file}" | awk -F'\t' '{print $1}')
			lock_process_time=$(cat "${XBSPSE_Crawler_lock_file}" | awk -F'\t' '{print $3}')
			lock_process_timeout=$(cat "${XBSPSE_Crawler_lock_file}" | awk -F'\t' '{print $4}')
			
			lock_process_delta_time=$((${current_time:-0} - ${lock_process_time:-0}))
			if [ ${lock_process_timeout:-0} -lt ${lock_process_delta_time:-0} ]; then
			    printf 'XBSPSE Crawler [%4d]: [%6d]> CODA DEL PROCESSO %s ELIMINATA\n' ${XBSPSE_Crawler_RANDOM} ${CRAWLER_TIMEOUT_COUNTDOWN} ${plugin_name} >&2
			    rm -f "${XBSPSE_Crawler_lock_file}"
			    sleep 1
			    CRAWLER_TIMEOUT_COUNTDOWN=$((${CRAWLER_TIMEOUT_COUNTDOWN} - 1))
			    
			    queue_connection_manager 'GET'
			    return $?
			elif [ ${flag_queue_locked_warning:-0} == 0 ]; then
			    flag_queue_locked_warning=1
			    printf 'XBSPSE Crawler [%4d]: [%6d]> CODA NON DISPONIBILE PER IL PROCESSO: %s\n' ${XBSPSE_Crawler_RANDOM} ${CRAWLER_TIMEOUT_COUNTDOWN} ${plugin_name} >&2
			fi
		    fi
		    
		    current_time=$(date +%s)
		done
		
		return $CODE_ERR_TIMEOUT
	    else
		printf 'XBSPSE Crawler [%4d]: CATTURA DELLA CODA PER IL PROCESSO: %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${plugin_name}" >&2
		
		printf '%d\t%s\t%d\t%d\t%s\t%d\t%s\t%s\t\n' \
		    "${XBSPSE_Crawler_RANDOM}" \
		    "${plugin_name}" \
		    "${current_time:-0}" \
		    "${CRAWLER_TIMEOUT_COUNTDOWN:-0}" \
		    "${irc_server_host}" \
		    "${irc_server_port}" \
		    "${irc_server_channel}" \
		    "${irc_nickname}" \
		    > "${XBSPSE_Crawler_lock_file}"
		
		sleep_random=$((${RANDOM:0:2} / 4))
		CRAWLER_TIMEOUT_COUNTDOWN=$((${CRAWLER_TIMEOUT_COUNTDOWN} - ${sleep_random}))
		[ ${CRAWLER_TIMEOUT_COUNTDOWN} -lt 0 ] && CRAWLER_TIMEOUT_COUNTDOWN=0
		sleep ${sleep_random}
		
		queue_connection_manager 'CHECK'
		if [ $? == 0 ]; then
		    return 0
		else
		    queue_connection_manager 'GET'
		fi
	    fi
	    ;;
	"RELEASE")
	    if [ -f "${XBSPSE_Crawler_lock_file}" ]; then
		queue_connection_manager 'CHECK'
		if [ $? == 0 ]; then
		    printf 'XBSPSE Crawler [%4d]: RILASCIO CODA PER IL PROCESSO: %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${plugin_name}" >&2
		    rm -f "${XBSPSE_Crawler_lock_file}"
		    return 0
		else
		    return $CODE_ERR_NO_QUEUE_LOCKED_FROM_THIS_PROCESS
		fi
	    else
		return $CODE_ERR_NO_QUEUE_LOCKED_FROM_THIS_PROCESS
	    fi
	    ;;
	"FORCE_RELEASE")
	    rm -f "${XBSPSE_Crawler_lock_file}"
	    ;;
	"RELEASE_PLUGIN")
	    [ -f "${XBSPSE_Crawler_lock_file}" ] && \
		[ "$(grep -i "	${2}	" "${XBSPSE_Crawler_lock_file}")" != "" ] && \
		return 1
	    
	    [ -f "${XBSPSE_Crawler_lock_file}" ] && rm -f "${XBSPSE_Crawler_lock_file}"
	    ;;
	"CHECK")
	    printf 'XBSPSE Crawler [%4d]: CONTROLLO CODA PER IL PROCESSO: %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${plugin_name}" >&2
	    [ -f "${XBSPSE_Crawler_lock_file}" ] && \
		[ "$(grep -i "^${XBSPSE_Crawler_RANDOM}	" "${XBSPSE_Crawler_lock_file}")" != "" ] && \
		return 0
	    
	    return $CODE_ERR_NO_QUEUE_LOCKED_FROM_THIS_PROCESS
	    ;;
	*)
	    printf 'XBSPSE Crawler [%4d]: ERRORE INTERNO\n' ${CRAWLER_TIMEOUT_COUNTDOWN} >&2
	    ;;
    esac
    
    return $CODE_ERR_TIMEOUT
}

#####################################################################
#____________________________________________________________________
function send_to_irc_server() {
    check_system_load_average
    
    printf "%s\r\n" "$*" >&3 2> /dev/null
    if [ $? != 0 ]; then
	printf 'XBSPSE Crawler [%4d]: RILEVATO PROBLEMA AD INVIARE IL COMANDO %s.\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "$*" >&2
	printf "ERR : %s\n" "$*" >> "${XBSPSE_Crawler_dat_file}"
    else
	printf "SEND: %s\n" "$*" >> "${XBSPSE_Crawler_dat_file}"
    fi
}

#####################################################################
#____________________________________________________________________
function connect_to_irc_server() {
    plugin_name=${1}
    irc_server_host=${2}
    irc_server_port=${3:-6667}
    irc_server_channel=${4:-#buduscript}
    irc_nickname=${5:-${USER}_${XBSPSE_Crawler_RANDOM}}
    irc_password=${6}
    
    CRAWLER_TIMEOUT_COUNTDOWN=${WGET_TIMEOUT:-16}
    queue_connection_manager 'GET'
    [ $? != 0 ] && return $?
    
    [ -z "${irc_server_host}" ] && return $CODE_ERR_NO_HOST
    [ -z "${irc_server_channel}" ] && return $CODE_ERR_NO_CHANNEL
    [ -f "${XBSPSE_Crawler_dat_file}" ] && \
	mv -f "${XBSPSE_Crawler_dat_file}" "${XBSPSE_Crawler_dat_file}.${plugin_name}" 2>> ${XBSPSE_Crawler_log_file} && \
	[ -f "${XBSPSE_Crawler_dat_file}" ] && \
	return $CODE_ERR_NO_DELETE_DAT_FILE
    
    printf 'XBSPSE Crawler [%4d]: CONNESSIONE AL SERVER IRC %s:%s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${irc_server_host}" "${irc_server_channel}" >&2
    {
	exec 3<>/dev/tcp/${irc_server_host}/${irc_server_port} 
    } 2>> ${XBSPSE_Crawler_log_file}
    [ ${?} != 0 ] && return $CODE_ERR_CONNECTION
    
    {
	while read irc_server_message; do
	    [[ ${irc_server_message:0:1} == ":" ]] && \
		irc_server_source="${irc_server_message%% *}" && \
		irc_server_message=$(echo "${irc_server_message#* }" | tr '\r' ' ')
	    
	    irc_server_source="${irc_server_source:1}"
	    irc_from_user=${irc_server_source%%\!*}
	    irc_message_text="${irc_server_message#*:}"
	    
	    if [ "$(echo "${irc_server_message}" | grep -iF 'VERSION')" != "" ]; then
		printf 'XBSPSE Crawler [%4d]: VERIFICA DELLA VERSIONE\n' ${CRAWLER_TIMEOUT_COUNTDOWN} >&2
		
		send_to_irc_server "NOTICE ${irc_from_user} :BuDuScRiPt XDCC Crawler $Revision: 1.24 $"
		send_to_irc_server "PRIVMSG ${irc_from_user} :BuDuScRiPt XDCC Crawler $Revision: 1.24 $"
	    fi
	    
	    if [ "$(echo "${irc_server_message}" | grep -iF 'ERROR :Closing Link:')" != "" ]; then
		printf 'XBSPSE Crawler [%4d]: CONNESSIONE CHIUSA\n' ${CRAWLER_TIMEOUT_COUNTDOWN} >&2
	    fi
	    
	    irc_server_type="${irc_server_message%% *}"
	    case ${irc_server_type} in
		"PING")
		    printf '%s\t%s\t%s\t%s\n' "${irc_server_type}" "${irc_server_source}" "${irc_from_user}" "${irc_server_message}"
		    
		    ping_code=$(echo "${irc_server_message}" | awk -F':' '{print $2}')
		    send_to_irc_server "PONG :${ping_code}"
		    
		    printf 'XBSPSE Crawler [%4d]: PONG: %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${ping_code}" >&2
		    if [ ! -f "${XBSPSE_Crawler_lock_file}" ]; then
			printf 'XBSPSE Crawler [%4d]: [%6d]> BLOCCO DELLA CODA PERSO\n' ${CRAWLER_TIMEOUT_COUNTDOWN} ${XBSPSE_Crawler_RANDOM} >&2
			send_to_irc_server "QUIT"
		    fi
		    ;;
		*)
		    printf '%s\t%s\t%s\t%s\n' "${irc_server_type}" "${irc_server_source}" "${irc_from_user}" "${irc_server_message}"
	    esac
	done
    } <&3 >> "${XBSPSE_Crawler_dat_file}" &

    [[ ${irc_password} ]] && send_to_irc_server "PASS ${irc_password}"
    send_to_irc_server "NICK ${irc_nickname}"
    send_to_irc_server "USER ${irc_nickname} localhost ${irc_server_host} :${irc_nickname}"
}

#####################################################################
#____________________________________________________________________
function find_irc_server_status() {
    if [ "${1}" != "" ]; then
	find_data=$(grep -iF "${1}" "${XBSPSE_Crawler_dat_file}" | grep -v '^SEND: ')
    fi
    if [ "${2}" != "" ]; then
	find_data=$(echo "${find_data}" | grep -iF "${2}")
    fi
    if [ "${3}" != "" ]; then
	find_data=$(echo "${find_data}" | grep -iF "${3}")
    fi
    if [ "${4}" != "" ]; then
	find_data=$(echo "${find_data}" | grep -iF "${4}")
    fi
    
    [ "${find_data}" != "" ] && return 1
    
    return 0
}

#####################################################################
#____________________________________________________________________
function find_irc_server_status_regexp() {
    if [ "${1}" != "" ]; then
	find_data=$(grep -i "${1}" "${XBSPSE_Crawler_dat_file}" | grep -v '^SEND: ')
    fi
    
    [ "${find_data}" != "" ] && return 1
    
    return 0
}

#####################################################################
#____________________________________________________________________
check_privmsg_to_bot_chance=0
function check_privmsg_to_bot() {
    check_privmsg_to_bot_chance=$((${check_privmsg_to_bot_chance} + 1))
    
    if [ ${check_privmsg_to_bot_chance:-0} -ge 3 ]; then
	printf 'XBSPSE Crawler [%4d]: TENTATIVO DI GESTIRE LA SITUAZIONE FALLITO, DEVO ABORTIRE.\n' ${CRAWLER_TIMEOUT_COUNTDOWN} >&2
	
	CRAWLER_TIMEOUT_COUNTDOWN=0
	
	return 0
    fi
    
    find_irc_server_status 'PRIVMSG	' "${irc_nickname}"
    if [ $? != 0 ]; then
	[ "${privmsg_nickname_quered}" == "" ] && \
	    printf 'XBSPSE Crawler [%4d]: IL CRAWLER HA RICEVUTO MESSAGGI PRIVATI; PROVO A RISPONDERE E RIMANERE IN ATTESA DELLA RISPOSA DESIDERATA.\n' ${CRAWLER_TIMEOUT_COUNTDOWN} >&2
	
	privmsg_nickname_list=$(echo "${find_data}" | awk -F'\t' '{print $3}' | sort | uniq)
	for privmsg_nickname in ${privmsg_nickname_list}; do
	    XBSPSE_NICKNAME_quered=$((${XBSPSE_NICKNAME_quered:-1} + 1))
	    
	    if [ "$(echo "${privmsg_nickname_quered}" | grep -iF "${privmsg_nickname}")" == "" ]; then
		printf 'XBSPSE Crawler [%4d]: PROVO AD INTERROGARE %s.\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${privmsg_nickname}" >&2
		
		send_to_irc_server "PRIVMSG ${privmsg_nickname} :${XBSPSE_STRING}" && \
		    printf 'XBSPSE Crawler [%4d]: INVIO DEL COMANDO    : %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "PRIVMSG ${privmsg_nickname} :${XBSPSE_STRING}" >&2
		
		privmsg_nickname_quered="${privmsg_nickname_quered} ${privmsg_nickname} "
	    else
		[ "$(echo "${privmsg_nickname_quered_msg}" | grep -iF "${privmsg_nickname_msg}")" == "" ] && \
		    printf 'XBSPSE Crawler [%4d]: EVITO DI INTERROGARE %s.\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${privmsg_nickname}" >&2 && \
		    privmsg_nickname_quered_msg="${privmsg_nickname_quered_msg} ${privmsg_nickname} "
	    fi
	done
	
	connection_status=3
    else
	connection_status=0
    fi
    
    return 1
}

#####################################################################
#____________________________________________________________________
function irc_connection_manager() {
    sleep_step_seconds=3

    irc_find_nickname=${1}
    GREETING_MSG=${2}
    XBSPSE_NICKNAME=${3:-${irc_server_channel}}
    XBSPSE_STRING=${4:-\!list}
    XBSPSE_FIND=${5}
    
    connection_status=0
    waiting_answer=0
    XBSPSE_NICKNAME_quered=0
    
    previous_connection_status=""
    
    while [ ${CRAWLER_TIMEOUT_COUNTDOWN} -gt 0 ] && \
	[ ${irc_connection_manager_loop:-0} == 0 ]; do
	
	if [ ! -f "${XBSPSE_Crawler_lock_file}" ]; then
	    printf 'XBSPSE Crawler [%4d]: [%6d]> BLOCCO DELLA CODA PERSO\n' ${CRAWLER_TIMEOUT_COUNTDOWN} ${XBSPSE_Crawler_RANDOM} >&2
	    
	    irc_connection_manager_loop=1
	    send_to_irc_server "QUIT"
	fi
	
	sleep ${sleep_step_seconds}
	
	if [ ${previous_connection_status:-X} != ${connection_status} ]; then
	    previous_connection_status=${connection_status}
	    
	    printf 'XBSPSE Crawler [%4d]: [%6d]> STATUS: %d\n' ${CRAWLER_TIMEOUT_COUNTDOWN} ${XBSPSE_Crawler_RANDOM} ${connection_status} >&2
	fi
	
	case ${connection_status} in
	    0)
		find_irc_server_status '001	'
		if [ $? != 0 ]; then
		    printf 'XBSPSE Crawler [%4d]: CONNESSO AL SERVER: %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${irc_server_host}" >&2
		    
		    send_to_irc_server "JOIN ${irc_server_channel}"
		    connection_status=1
		fi
		
		flag_already_connected=$((${flag_already_connected:-0} + 1))
		if [ ${flag_already_connected} -gt 10 ]; then
		    printf 'XBSPSE Crawler [%4d]: ATTENZIONE: IL SERVER POTREBBE IMPEDIRE CONNESSIONI MULTIPLE, VERIFICARE CHE XCHAT NON SIA COLLEGATO AL SERVER %s.\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${irc_server_host}" >&2
		fi
		if [ ${flag_already_connected} -gt 20 ]; then
		    printf 'XBSPSE Crawler [%4d]: ATTENZIONE: SEMBRA NON SIA POSSIBILE COLLEGARSI AL SERVER %s. DEVO ABORTIRE.\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${irc_server_host}" >&2
		    
		    CRAWLER_TIMEOUT_COUNTDOWN=${sleep_step_seconds}
		fi
		;;
	    1)
		find_irc_server_status 'JOIN	' "	${irc_nickname}	" "JOIN :${irc_server_channel} "
		[ $? == 0 ] && find_irc_server_status 'JOIN	' "	${irc_nickname}	" "JOIN ${irc_server_channel} "
		if [ $? != 0 ]; then
		    printf 'XBSPSE Crawler [%4d]: JOIN DEL CANALE      : %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${irc_server_channel}" >&2
		    
		    IFS_OLD=${IFS}
		    IFS=$'\n'
		    for SINGLE_GREETING_MSG in ${GREETING_MSG}; do
			printf 'XBSPSE Crawler [%4d]: INVIO DEL SALUTO     : %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${SINGLE_GREETING_MSG}" >&2
		    
			send_to_irc_server "PRIVMSG ${irc_server_channel} :${SINGLE_GREETING_MSG}"
			sleep 4
		    done
		    IFS=${IFS_OLD}
		    connection_status=2
		fi
		
		flag_already_joined=$((${flag_already_joined:-0} + 1))
		if [ ${flag_already_joined} -gt 9 ]; then
		    printf 'XBSPSE Crawler [%4d]: ATTENZIONE: IL CANALE %s POTREBBE IMPEDIRE IL JOIN DEL CRAWLER.\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${irc_server_channel}" >&2
		    
		    check_privmsg_to_bot
		fi
		[ ${flag_already_joined} -gt 10 ] && CRAWLER_TIMEOUT_COUNTDOWN=${sleep_step_seconds}
		;;
	    2)
		find_irc_server_status '366	'
		if [ $? != 0 ]; then
		    for XBSPSE_NICKNAME_single in ${XBSPSE_NICKNAME}; do
			XBSPSE_NICKNAME_quered=$((${XBSPSE_NICKNAME_quered:-0} + 1))
			printf 'XBSPSE Crawler [%4d]: INVIO DEL COMANDO    : %s [%d]\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "PRIVMSG ${XBSPSE_NICKNAME_single} :${XBSPSE_STRING}" ${XBSPSE_NICKNAME_quered} >&2
			
			send_to_irc_server "PRIVMSG ${XBSPSE_NICKNAME_single} :${XBSPSE_STRING}"
		    done
		    connection_status=3
		else
		    printf 'XBSPSE Crawler [%4d]: ATTESA RICEZIONE LISTA UTENTI\n' ${CRAWLER_TIMEOUT_COUNTDOWN} >&2
		fi
		;;
	    3)
		waiting_loop_counter=$((${waiting_loop_counter:-0} + 1))
		
		printf 'XBSPSE Crawler [%4d]: ATTESA RISPOSTA [%d]   : %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} ${waiting_loop_counter:-0} "${XBSPSE_FIND:-"\!"list}" >&2
		
		find_irc_server_status_regexp "${XBSPSE_FIND}"
		[ $? != 0 ] && \
		    printf '%s\n' "${find_data}" && \
		    irc_connection_manager_loop=1 && \
		    printf 'XBSPSE Crawler [%4d]: INTERCETTATA RISPOSTA: %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${find_data}" >&2
		
		[ ${waiting_loop_counter:-0} -gt 3 ] && \
		    connection_status=4
		
		[ ${waiting_loop_counter:-0} -gt 1 ] && \
		    printf 'XBSPSE Crawler [%4d]: INVIO DEL COMANDO    : %s [%d]\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "PRIVMSG ${XBSPSE_NICKNAME_single} :${XBSPSE_STRING}" ${XBSPSE_NICKNAME_quered} >&2 && \
		    send_to_irc_server "PRIVMSG ${XBSPSE_NICKNAME_single} :${XBSPSE_STRING}"
		;;
	    4)
		find_irc_server_status_regexp "${XBSPSE_FIND}"
		[ $? != 0 ] && \
		    printf '%s\n' "${find_data}" && \
		    irc_connection_manager_loop=1 && \
		    printf 'XBSPSE Crawler [%4d]: INTERCETTATA RISPOSTA: %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${find_data}" >&2
		
		for XBSPSE_NICKNAME_single in ${XBSPSE_NICKNAME}; do
		    find_irc_server_status '401	' "${XBSPSE_NICKNAME_single} :"
		    if [ $? != 0 ]; then
			XBSPSE_NICKNAME_quered=$((${XBSPSE_NICKNAME_quered:-1} - 1))
			
			[ ${XBSPSE_NICKNAME_quered:-0} -ge 0 ] && \
			    printf 'XBSPSE Crawler [%4d]: ATTENZIONE: IL BOT %s NON RISULTA PRESENTE NEL CANALE %s [%d].\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${XBSPSE_NICKNAME_single}" "${irc_server_channel}" ${XBSPSE_NICKNAME_quered:-0} >&2
			
			if [ ${XBSPSE_NICKNAME_quered:-0} -le 0 ]; then
			    [ "${privmsg_nickname_quered}" == "" ] && \
				check_privmsg_to_bot
			    
			    if [ ${connection_status:-0} == 0 ]; then
				printf 'XBSPSE Crawler [%4d]: ATTENZIONE: TENTATIVO DI RISOLVERE LA MANCANZA DEL BOT %s NEL CANALE %s FALLITA; PROVO DACCAPO PER LA %d VOLTA.\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${XBSPSE_NICKNAME_single}" "${irc_server_channel}" "${check_privmsg_to_bot_chance}" >&2
				send_to_irc_server "PRIVMSG ${irc_server_channel} :${LIST_COMMAND}" && \
				    printf 'XBSPSE Crawler [%4d]: INVIO DEL COMANDO    : %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "PRIVMSG ${irc_server_channel} :${LIST_COMMAND}" >&2
			    else
				privmsg_nickname_quered=""
			    fi
			else
			    printf 'XBSPSE Crawler [%4d]: ATTENZIONE: IL BOT %s NON RISULTA PRESENTE NEL CANALE %s ; VERIFICO GLI ALTRI %d.\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${XBSPSE_NICKNAME_single}" "${irc_server_channel}" "${XBSPSE_NICKNAME_quered}" >&2
			fi
		    else
			if [ ${waiting_answer} -ge 3 ]; then
			    printf 'XBSPSE Crawler [%4d]: ATTENZIONE: IL BOT %s RISULTA PRESENTE NEL CANALE %s MA NON RISPONDE, DEVO ABORTIRE.\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "${XBSPSE_NICKNAME_single}" "${irc_server_channel}" >&2
			    
			    CRAWLER_TIMEOUT_COUNTDOWN=0
			elif [ ${waiting_answer} -ge 1 ]; then
			    printf 'XBSPSE Crawler [%4d]: INVIO DEL COMANDO    : %s\n' ${CRAWLER_TIMEOUT_COUNTDOWN} "PRIVMSG ${XBSPSE_NICKNAME_single} :${XBSPSE_STRING}" >&2
			    
			    send_to_irc_server "PRIVMSG ${XBSPSE_NICKNAME_single} :${XBSPSE_STRING}"
			fi
			waiting_answer=$((${waiting_answer} + 1))
		    fi
		done
		;;
	    *)
		printf 'XBSPSE Crawler [%4d]: ERRORE INTERNO.\n' ${CRAWLER_TIMEOUT_COUNTDOWN} >&2
	esac
	
	CRAWLER_TIMEOUT_COUNTDOWN=$((${CRAWLER_TIMEOUT_COUNTDOWN} - ${sleep_step_seconds}))
	[ ${CRAWLER_TIMEOUT_COUNTDOWN} -lt 0 ] && CRAWLER_TIMEOUT_COUNTDOWN=0
    done
    
    printf 'XBSPSE Crawler [%4d]: INVIO COMANDO QUIT\n' ${CRAWLER_TIMEOUT_COUNTDOWN} >&2
    send_to_irc_server "QUIT"
    
    [ -f /tmp/XBSPSE_Crawler_file.dat ] && rm -f /tmp/XBSPSE_Crawler_file.dat
    mv -f "${XBSPSE_Crawler_dat_file}" /tmp/XBSPSE_Crawler_file.dat
    
    printf '%d\n' ${CRAWLER_TIMEOUT_COUNTDOWN} > "${XBSPSE_Crawler_time_file}"
    
    queue_connection_manager 'RELEASE'
    [ $? != 0 ] && return $?
    
    return ${irc_connection_manager_loop};
}

#####################################################################
#____________________________________________________________________
function irc_connection_error() {
    connect_to_irc_server_status=${?}
    
    case "${connect_to_irc_server_status}" in
	"$CODE_ERR_NO_HOST")
	    printf '[XBSPSE Crawler]\tERROR: No IRC Host Specified.\n' >&2
	    ;;
	"$CODE_ERR_NO_CHANNEL")
	    printf '[XBSPSE Crawler]\tERROR: No IRC Channel Specified.\n' >&2
	    ;;
	"$CODE_ERR_CONNECTION")
	    printf '[XBSPSE Crawler]\tERROR: IRC Server Connection; See log for more info: %s\n' "${XBSPSE_Crawler_log_file}" >&2
	    ;;
	"$CODE_ERR_NO_DELETE_DAT_FILE")
	    printf '[XBSPSE Crawler]\tERROR: Database Problem; See log for more info: %s\n' "${XBSPSE_Crawler_log_file}" >&2
	    ;;
	"$CODE_ERR_TIMEOUT")
	    printf '[XBSPSE Crawler]\tERROR: Reached TIMEOUT, ABORT.\n' >&2
	    ;;
	"$CODE_ERR_LOST_LOCK_ON_QUEUE")
	    printf '[XBSPSE Crawler]\tERROR: Lost Queue Lock.\n' >&2
	    ;;
	*)
	    return 0
	    ;;
    esac
    
    return ${connect_to_irc_server_status}
}

#####################################################################
#____________________________________________________________________
function get_XBSPSE_ID() {
    printf '%d\n' ${XBSPSE_Crawler_RANDOM}
}

#####################################################################
#____________________________________________________________________
function get_XBSPSE_TIMEOUT() {
    XBSPSE_Crawler_time_file=/tmp/XBSPSE_Crawler_${1}.time
    
    [ -f "${XBSPSE_Crawler_time_file}" ] && \
	printf '%d\n' $(cat "${XBSPSE_Crawler_time_file}") && \
	rm -f "${XBSPSE_Crawler_time_file}"
}

#####################################################################
#____________________________________________________________________
# MAIN DI TEST
if [ ${CRAWLER_TEST:-NO} != "NO" ]; then
    WGET_TIMEOUT=60
    
    connect_to_irc_server "CRAWLER_TEST" "${CRAWLER_SERVER:-192.168.0.1}" "${CRAWLER_PORT:-6667}" "${CRAWLER_CHANNEL:-#testing}"
    irc_connection_error
    [ $? == 0 ] && \
	irc_connection_manager "BOTSERV" "Hello World" "BOTSERV" "Scrivi: TEST OK" "TEST OK"
fi

#####################################################################
#____________________________________________________________________